//grbl
#define GRBL_PWM_IN            3
#define GRBL_INDEX_OUT         9
#define GRBL_SYNC_OUT          10
#define GRBL_ENABLE_SPINDLE_IN 13
//motor encoder
#define MOTOR_ENCODER_IN       2
//motor driver
#define MOTOR_ENABLE_OUT       5
#define MOTOR_PWM_OUT          6

#define MOTOR_MIN_SPEED        150
#define MOTOR_MAX_SPEED        250

void GRBL_PWM_RISING();
void GRBL_PWM_FALLING();
void MOTOR_ENCODER();

unsigned int GRBL_Time_ms = 0;
unsigned int GRBL_PWM_TEMP = 0;
float GRBL_PWM_VALUE = 0;

unsigned int MOTOR_SPEED = 0;
bool         MOTOR_REFRESH_PID = false;
unsigned int MOTOR_SPEED_TEMP = 0;
unsigned int MOTOR_INDEX_TEMP = 0;
unsigned int MOTOR_SYNC_TEMP = 0;
unsigned int MOTOR_PWM_TEMP = 0;

unsigned int PWM_OUT_FROM_PID = 0; //cycles of timer 171Hz

void setup() {
  pinMode(GRBL_PWM_IN, INPUT);
  pinMode(GRBL_INDEX_OUT, OUTPUT);
  pinMode(GRBL_SYNC_OUT, OUTPUT);
  pinMode(GRBL_ENABLE_SPINDLE_IN, INPUT);

  pinMode(MOTOR_ENCODER_IN, INPUT);

  pinMode(MOTOR_ENABLE_OUT, OUTPUT);
  pinMode(MOTOR_PWM_OUT, OUTPUT);
  
  attachInterrupt(digitalPinToInterrupt(GRBL_PWM_IN), GRBL_PWM_RISING, RISING );
  attachInterrupt(digitalPinToInterrupt(GRBL_PWM_IN), GRBL_PWM_FALLING, FALLING );
  attachInterrupt(digitalPinToInterrupt(MOTOR_ENCODER_IN), MOTOR_ENCODER, RISING  );
  //PID loop clock 171Hz
  TCCR1B = 0;
  OCR1A = 365;
  TCCR1B = (1<<WGM12) | (1<<CS12); 
  TIMSK1 = (1<<OCIE1A);
  //PWM
  OCR0A = PWM_OUT_FROM_PID;
  TCCR0A = (1 << COM0A1) | (1<<WGM01) | (1<<WGM00); 
  TCCR0B = (1 << CS00);
}

bool Writed = false;
float Kp = 1.0;
float Ki = 1.0;
float Kd = 1.0;
float error_sum = 1.0;
float temp = 0.0;
float error = 0.0;

void loop()
{
  if( MOTOR_REFRESH_PID )//171Hz
  {
    if( digitalRead( GRBL_ENABLE_SPINDLE_IN ) == LOW )
    {
      digitalWrite(MOTOR_ENABLE_OUT, LOW);
      PWM_OUT_FROM_PID = 0;
      Writed = false;
    }
    else
    {
      if( Writed == false )
      {
        digitalWrite(MOTOR_ENABLE_OUT, HIGH);
        Writed = true;
      }
      //refresh every 171 times per second
      //PID controler
      // GRBL_PWM_VALUE : target speed 0-100%
      // PWM_OUT_FROM_PID : output PWM 0-100%
      // MOTOR_SPEED : real motor speed 0-100 impulses per 1/171s
      error = (GRBL_PWM_VALUE/100.0F) - (MOTOR_SPEED/100.0F);
      error_sum += error;

      float PID = (Kp * error) + (Ki * error_sum) + (Kd * (error - temp));
      temp = error;
      PWM_OUT_FROM_PID = (int)(PID *255.0F);
      if( PWM_OUT_FROM_PID < MOTOR_MIN_SPEED )//minimum speed
        PWM_OUT_FROM_PID = MOTOR_MIN_SPEED;
      if( PWM_OUT_FROM_PID > MOTOR_MAX_SPEED )
        PWM_OUT_FROM_PID = MOTOR_MAX_SPEED;
      //test half speed
      //PWM_OUT_FROM_PID = MOTOR_MIN_SPEED + (MOTOR_MAX_SPEED-MOTOR_MIN_SPEED)/2;
      OCR0A = PWM_OUT_FROM_PID;
      MOTOR_REFRESH_PID = false;
    }
  }
}

ISR(TIMER1_COMPA_vect) //171Hz
{
  MOTOR_SPEED = (MOTOR_SPEED_TEMP*100)/374;
  MOTOR_SPEED_TEMP = 0;
  MOTOR_REFRESH_PID = true;
}

void MOTOR_ENCODER()
{
  MOTOR_SPEED_TEMP++;
  MOTOR_INDEX_TEMP++;
  MOTOR_SYNC_TEMP++;
  if( MOTOR_SYNC_TEMP == 234 )//1/8 rotation
  {
    MOTOR_SYNC_TEMP = 0;
    digitalWrite(GRBL_SYNC_OUT, HIGH);
  }
  if( MOTOR_INDEX_TEMP == 1875 )//full rotation
  {
    MOTOR_INDEX_TEMP = 0;
    digitalWrite(GRBL_INDEX_OUT, HIGH);
  }
  if( MOTOR_SYNC_TEMP == 10 )
    digitalWrite(GRBL_SYNC_OUT, LOW);
  if( MOTOR_INDEX_TEMP == 10 )
    digitalWrite(GRBL_INDEX_OUT, LOW);
}

void GRBL_PWM_RISING()
{
  if( GRBL_PWM_TEMP > 0 )
    GRBL_PWM_VALUE = ((float)GRBL_PWM_TEMP / (float)(micros()-GRBL_Time_ms))*100;
  GRBL_Time_ms = micros();
}

void GRBL_PWM_FALLING()
{
  GRBL_PWM_TEMP = micros()-GRBL_Time_ms;
  GRBL_Time_ms = micros();
}
